////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "GLES20TriangleRenderer.h"

#include <gtf/scdr/sdk/GraphicsContext.h>
#include <gtf/scdr/sdk/GraphicsExecutionContext.h>
#include <gtf/scdr/sdk/WidgetUtils.h>

#include <GtfTypes/GtfMacros.h>

#include <string.h>
#include <stdio.h>
#include <new>

static const char* VertexShader =
"uniform highp mat4 uProjMat;                          \n" \
"uniform highp mat4 uModelMat;                         \n" \
"uniform lowp vec4 uColor;                             \n" \
"                                                      \n" \
"attribute highp vec2 aPosition;                       \n" \
"                                                      \n" \
"varying lowp vec4 vColor;                             \n" \
"                                                      \n" \
"void main()                                           \n" \
"{                                                     \n" \
"    vColor = uColor;                                  \n" \
"    vec4 pos = uModelMat * vec4(aPosition, 0.0, 1.0); \n" \
"    gl_Position = uProjMat * pos;                     \n" \
"}                                                     \n";

static const char* FragmentShader =
"varying lowp vec4 vColor;                             \n" \
"                                                      \n" \
"void main()                                           \n" \
"{                                                     \n" \
"    gl_FragColor = vColor;                            \n" \
"}                                                     \n";

uint16_t GLES20TriangleRenderer::colorHandle = 0;

// Callback for instantiating a new renderer. This callback is set in 
// TrianglePlugin::registerRenderer
gtf::scdr::sdk::WidgetRenderer* GLES20TriangleRenderer::createIntance(
    const gtf::scdr::sdk::WidgetRendererRegistration& registration_,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    const char* names[] = { "Color" };
    
    // For accessing the properties uint16_t handles are used.
    // Resolving the handles by the SCDR name of the properties once at startup.
    registration_.getPropertyHandles(names, 1, &colorHandle);

    // Allocate memory for the custom renderer by using the SCDR memory management.
    void* mem = context_.alloc(sizeof(GLES20TriangleRenderer));
    // Use placement new to construct the renderer.
    GLES20TriangleRenderer* glesRenderer = new (mem)GLES20TriangleRenderer();
    if (NULL == glesRenderer)
    {
        return NULL;
    }

    // Initialize the custom renderer
    if (!glesRenderer->init(context_))
    {
        // Cleanup on failure
        GLES20TriangleRenderer::destroyInstance(glesRenderer, registration_, context_, NULL);
        glesRenderer = NULL;
    }

    return glesRenderer;
}

// Callback for destroying a renderer. This callback is set in 
// TrianglePlugin::registerRenderer
void GLES20TriangleRenderer::destroyInstance(
    gtf::scdr::sdk::WidgetRenderer* renderer_,
    const gtf::scdr::sdk::WidgetRendererRegistration&,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    GTF_OSAL_ASSERT(renderer_);
    if (renderer_)
    {
        // Downcast the renderer to a pointer of the custom renderer.
        GLES20TriangleRenderer* renderer = static_cast<GLES20TriangleRenderer*>(renderer_);

        // De-initialize the renderer
        renderer->deinit(context_);

        // Call the destructor manually because the memory deallocation does not take care of object destruction.
        renderer->~GLES20TriangleRenderer();

        // Free the renderer using the SCDR memory management
        context_.free(renderer_);
    }
}

GLES20TriangleRenderer::GLES20TriangleRenderer()
    :gtf::scdr::sdk::WidgetRenderer()
    ,renderListHead(NULL)
    ,renderListTrail(NULL)
{
    memset(&userData.shader, 0, sizeof(ShaderParams));
}

GLES20TriangleRenderer::~GLES20TriangleRenderer()
{
}

bool GLES20TriangleRenderer::init(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The init method is called once at the instantiation of the renderer. 
    // Use the init method to initialize all graphics resources which do not
    // change during the lifetime of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer. 
    // executeUpdateFunction shall be used to initialize graphics resources.
    // The parameter doInit is a pointer to a function implementing the custom specific
    // OpenGL code for initialization. User data is a pointer passed to the doInit
    // function.
    return context_.executeUpdateFunction(doInit, &userData);
}

void GLES20TriangleRenderer::deinit(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The deinit method is called once at the destruction of the renderer.
    // Use the deinit method to destroy graphics resources at the end of the lifetime 
    // of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer.
    // executeDeleteFunction shall be used to destroy graphics resources.
    // The parameter doDeinit is a pointer to a function implementing the custom specific
    // OpenGL code for de-initialization. User data is a pointer passed to the doDeinit
    // function.
    context_.executeDeleteFunction(doDeinit, &userData);
}

bool GLES20TriangleRenderer::create(void* renderData_, gtf::scdr::sdk::GraphicsContext&)
{
    // The create method is called once for every widget instance created in the system. 
    // Use this method to do initialization tasks per widget.
    // The pointer renderData points to the user specific data of this widget instance.
    // The memory referenced by render data is provided by the SCDR. The size of the memory block 
    // is configured in TrianglePlugin::registerRenderer. It has the size of 
    // GLES20TriangleRenderer::RenderDataSize, which is the size of RenderData.
    // Do not free this memory, it is managed by the SCDR.
    RenderData* renderData = static_cast<RenderData*>(renderData_);
    memset(renderData, 0, sizeof(RenderData));
    return true;
}

bool GLES20TriangleRenderer::destroy(void* renderData_, gtf::scdr::sdk::GraphicsContext&)
{
    // The destroy method is called once when an instance of a widget is destroyed. 
    // Use this method to do de-initialization tasks per widget.
    memset(renderData_, 0, sizeof(RenderData));
    return true;
}

bool GLES20TriangleRenderer::begin(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the beginning of each frame rendered by the SCDR.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;  
    renderListTrail = NULL;
    return true;
}

int32_t GLES20TriangleRenderer::traverse(void* renderData_, gtf::scdr::sdk::Widget& widget_, gtf::scdr::sdk::TraverserContext&, gtf::scdr::sdk::GraphicsContext&)
{
    // When the widget is invisible, so are the children, hence abort traversal
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::IsWidgetVisible(widget_));

    GTF_OSAL_ASSERT(NULL != renderData_);

    // Cast the user memory to the RenderData structure
    RenderData* renderData = static_cast<RenderData*>(renderData_);
    gtf::scdr::sdk::WidgetDefaultProperties& properties = renderData->properties;

    // Load the default values in the SystemWidgetPropertySets::Full2D
    // Note that the property set Full2D shall be set in TrianglePlugin::registerRenderer
    // to make the default features work correctly.
    // This is an auxiliary function to cover the simple case, no properties handed over to the 
    // child widgets are modified.
    // An example for using the get and set functions for every single property 
    // is shown in the ripple effect example in GLES20RippleEffect::traverse.
    gtf::scdr::sdk::GetWidgetDefaultProperties(widget_, properties);

    // Set the transformation matrix of this widget, add scaling and z-translation
    renderData->modelMat = properties.widgetTransformationMatrix * 
        gtf::math::float4x4::Translation(0.0f, 0.0f, static_cast<float>(widget_.getZOrder())) *
        gtf::math::float4x4::Scaling2D(static_cast<float>(properties.width), static_cast<float>(properties.height));

    // Set the depth (drawId), used for scheduling of the render calls
    renderData->drawId = static_cast<uint16_t>((widget_.getZOrder()));

    uint32_t uintColor = 0;
    // Get custom property color. In GTF a color is represented as uint32 RGBA value.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(widget_.getValue(colorHandle, GtfTypeId::eTypeId_color, 1, &uintColor, sizeof(uintColor)));
    gtf::scdr::sdk::Color color = gtf::scdr::sdk::UnpackColor(uintColor);

    // Apply the coloration color to the widgets fill color
    color = gtf::scdr::sdk::ColorMultiply(color, properties.coloration);
    memcpy(renderData->fillColor, &color, sizeof(gtf::scdr::sdk::Color));

    // Append the render data of this widget to the linked list.
    renderListAppend(renderData);

    return TRAVERSE_SUCCESS;
}

bool GLES20TriangleRenderer::update(gtf::scdr::sdk::GraphicsContext&)
{
    // Here is typically code issued which updates graphics resources like
    // VBOs or textures. Graphics code for updating this resources would be issued 
    // with a GraphicsContext::executeUpdateFunction command.
    // Nothing to do for the triangle widget here.
    return true;
}

bool GLES20TriangleRenderer::render(gtf::scdr::sdk::GraphicsContext& context_)
{
    // Walk over the linked list with the render data and issue the rendering for each object
    for (RenderData* renderData = renderListHead; renderData; renderData = renderData->next)
    {
        // Use the SCDR allocFrame to create a RenderFunctionContext structure. 
        RenderFunctionContext* renderFunctionContext = static_cast<RenderFunctionContext*>(context_.allocFrame(sizeof(RenderFunctionContext)));

        if (NULL == renderFunctionContext)
        {
            // Failed to allocate memory. Stop the rendering and return an error.
            return false;
        }

        // For rendering the per widget render data (RenderData) and the common data 
        // for all instances (UserData) are needed.
        renderFunctionContext->renderData = renderData;
        renderFunctionContext->userData = &userData;

        // Determine whether the widget is opaque.
        const bool opaque = isOpaque(renderData->fillColor);

        // Generate a scheduler key. The SCDR ensures that all internal and external widgets
        // are rendered in a correct and fast order. The data in the schedule key is used for
        // sorting the render jobs.
        gtf::scdr::sdk::CScheduleDescriptor schedule(renderData->properties.framebufferId, renderData->drawId, opaque);

        // Issue a the delete function doDelete for the RenderFunctionContext structure.
        // Use the executeDeleteFunction to ensure that the renderFunctionContext remains 
        // alive during rendering executed by executeRenderFunction.
        if (false == context_.executeDeleteFunction(doDelete, renderFunctionContext))
        {
            // In case the execute delete function was not issued successfully
            // free the memory in place and return an error.
            context_.freeFrame(renderFunctionContext);
            return false;
        }

        // Execute rendering by issuing the executeRenderFunction. 
        // Parameters are 
        // * the doRender function, which implements the rendering
        // * the renderFunctionContext structure with the render data
        // * the schedule structure with the schedule data.
        context_.executeRenderFunction(doRender, renderFunctionContext, schedule);
    }

    return true;
}

bool GLES20TriangleRenderer::end(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the end of each frame rendered.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

bool GLES20TriangleRenderer::doInit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext&)
{
    // Implementation of the initialization code of the renderer.
    // This method is issued by GLES20TriangleRenderer::init. Since 
    // userData_ is a pointer to UserData, set in GLES20TriangleRenderer::init,
    // the void pointer can be casted here.
    if (NULL == userData_)
    {
        return false;
    }

    UserData* data = static_cast<UserData*>(userData_);
    return initShader(data->shader, VertexShader, FragmentShader);
}

bool GLES20TriangleRenderer::doRender(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the render code. This code was issued by GLES20TriangleRenderer::render
    // once for each widget to render.
    if (NULL == userData_)
    {
        return false;
    }

    // Cast the user data to the user specific data structures.
    RenderFunctionContext* ctx = static_cast<RenderFunctionContext*>(userData_);
    const UserData* userData = ctx->userData;
    const RenderData* renderData = ctx->renderData;

    if (NULL == userData || NULL == renderData)
    {
        return false;
    }

    // Set the framebuffer to be used for this widget in the render context.
    context_.setFramebuffer(renderData->properties.framebuffer);

    static const float vertices[] =
    {
        0.5f, 1.0f,
        0.0f, 0.0f,
        1.0f, 0.0f
    };

    // OpenGL|ES 2.0 implementation of the rendering

    // Disable some render states. Potentially every state can be changed by the SCDR,
    // in every frame. Before rendering the states shall be set to proper values.
    glDisable(GL_SCISSOR_TEST);
    glDisable(GL_STENCIL_TEST);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glDisable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    // Blending
    if (isOpaque(renderData->fillColor))
    {
        glDisable(GL_BLEND);
    }
    else
    {
        glEnable(GL_BLEND);
        glBlendEquationSeparate(GL_FUNC_ADD, GL_FUNC_ADD);
        glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
        glBlendColor(0.0f, 0.0f, 0.0f, 0.0f);
    }

    // It is also possible to just always use the scissor test but use the default max scissor rectangle in case there is 
    // nothing to scissor away.
    if (hasScissor(renderData->properties.scissor))
    {
        // OpenGL|ES 2.0 expects the (0/0) coordinate at the bottom left corner and x, y, width, height as parameter for the scissor
        // EB GUIDE Studio has (0/0) at the top left corner and xmin, ymin, xmax, ymax as parameter for the scissor
        // To set the scissor rectangle properly the coordinates are converted.
        int16_t scissor[4];
        const int16_t* const scissorIn = renderData->properties.scissor;
        const int16_t* const viewportIn = renderData->properties.viewport;

        scissor[0] = scissorIn[0];
        scissor[2] = scissorIn[2] - scissorIn[0]; // calculate the width
        scissor[3] = scissorIn[3] - scissorIn[1]; // calculate the height
        scissor[1] = static_cast<int16_t>(viewportIn[3]) - (scissorIn[1] + scissor[3]); // Flip the y coordinate: scissor.y = viewport.height - (scissor.y + scissor.height)

        glEnable(GL_SCISSOR_TEST);
        glScissor(scissor[0], scissor[1], scissor[2], scissor[3]);
    }
    else
    {
        glDisable(GL_SCISSOR_TEST);
    }

    glUseProgram(userData->shader.program);

    GLint vertexLocation = glGetAttribLocation(userData->shader.program, "aPosition");

    glVertexAttribPointer(vertexLocation, 2, GL_FLOAT, GL_FALSE, 2 * sizeof(float), vertices);
    glEnableVertexAttribArray(vertexLocation);

    // The viewport is only correct as long as it has the same size as the screen. 
    // If this is not the case, the different origin of OpenGL and GTF has to be considered.
    glViewport(renderData->properties.viewport[0], renderData->properties.viewport[1], renderData->properties.viewport[2], renderData->properties.viewport[3]);

    glUniformMatrix4fv(userData->shader.projMatLocation, 1, GL_FALSE, renderData->properties.viewProjectionMatrix.values);
    glUniformMatrix4fv(userData->shader.modelMatLocation, 1, GL_FALSE, renderData->modelMat.values);
    glUniform4fv(userData->shader.colorLocation, 1, renderData->fillColor);


    glDrawArrays(GL_TRIANGLES, 0, 3);

    return (GL_NO_ERROR == glGetError());
}

bool GLES20TriangleRenderer::doDeinit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext&)
{
    if (NULL == userData_)
    {
        return false;
    }

    // Implementation of the deinit code. This code is issued by GLES20TriangleRenderer::deinit
    UserData* data = static_cast<UserData*>(userData_);

    if (data->shader.program && data->shader.fragmentShader) glDetachShader(data->shader.program, data->shader.fragmentShader);
    if (data->shader.program && data->shader.vertexShader) glDetachShader(data->shader.program, data->shader.vertexShader);
    if (data->shader.program) glDeleteProgram(data->shader.program);
    if (data->shader.fragmentShader)glDeleteShader(data->shader.fragmentShader);
    if (data->shader.vertexShader) glDeleteShader(data->shader.vertexShader);

    data->shader.program = 0;
    data->shader.vertexShader = 0;
    data->shader.fragmentShader = 0;

    return true;
}

bool GLES20TriangleRenderer::doDelete(void* memory_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the doDelete. This code is issued by GLES20TriangleRenderer::render to free the 
    // RenderFunctionContext structure
    context_.freeFrame(memory_);
    return true;
}

void GLES20TriangleRenderer::renderListAppend(RenderData* data_)
{
    // Append data_ to the internal linked list with the user data.
    (NULL == renderListHead) ? (renderListHead = data_) : (renderListTrail->next = data_);
    renderListTrail = data_;
}

bool GLES20TriangleRenderer::initShader(ShaderParams& params_, const char* vertexShader_, const char* fragmentShader_)
{
    // Create and upload a shader program to the GLES20 graphics API
    GLuint vertexShaderHandle = loadShader(GL_VERTEX_SHADER, vertexShader_);
    GLuint fragmentShaderHandle = loadShader(GL_FRAGMENT_SHADER, fragmentShader_);

    GLuint program = glCreateProgram();

    GLint linked = 0;

    if ((0 == vertexShaderHandle) || (0 == fragmentShaderHandle) || (0 == program))
    {
        return false;
    }

    glAttachShader(program, vertexShaderHandle);
    glAttachShader(program, fragmentShaderHandle);

    glBindAttribLocation(program, 0, "aPosition");
    glLinkProgram(program);

    glGetProgramiv(program, GL_LINK_STATUS, &linked);

    if (0 == linked)
    {
        char error[1024];
        glGetProgramInfoLog(program, 1024, NULL, error);
        fprintf(stderr, "%s\n", error);
    }

    params_.projMatLocation = glGetUniformLocation(program, "uProjMat");
    params_.modelMatLocation = glGetUniformLocation(program, "uModelMat");
    params_.colorLocation = glGetUniformLocation(program, "uColor");

    params_.program = program;
    params_.vertexShader = vertexShaderHandle;
    params_.fragmentShader = fragmentShaderHandle;

    return (GL_NO_ERROR == glGetError());
}

GLuint GLES20TriangleRenderer::loadShader(GLenum type_, const char* shaderSrc_)
{
    // Create and upload a shader to the GLES20 graphics API
    GLint compiled = 0;
    GLuint shader = glCreateShader(type_);

    if (0 == shader)
    {
        return 0;
    }

    glShaderSource(shader, 1, &shaderSrc_, NULL);
    glCompileShader(shader);
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled);

    if (0 == compiled)
    {
        char error[1024];
        glGetShaderInfoLog(shader, sizeof(error), NULL, error);
        fprintf(stderr, "%s\n", error);
    }

    return shader;
}

bool GLES20TriangleRenderer::hasScissor(const int16_t scissorRect_[])
{
    return (scissorRect_[0] != 0 || 
            scissorRect_[1] != 0 ||
            scissorRect_[2] != GtfLimits::int16_max || 
            scissorRect_[3] != GtfLimits::int16_max);
}

bool GLES20TriangleRenderer::isOpaque(const float color_[])
{
    return (color_[3] == 1.0);
}