////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <GtfPluginLoader/GtfPluginSymbols.h>
#include <GtfPluginLoader/GtfPluginLoader.h>
#include <GtfPluginLoader/GtfPluginFactory.h>
#include <GtfRunlevel/GtfRunlevel.h>
#include <GtfCoreRuntime/GtfCoreRuntime.h>
#include <GtfType/GtfTypeManager.h>
#include <GtfPluginLoader/GtfPluginSymbols.h>
#include <GtfMetaInfo/DescriptorProvider.h>
#include <GtfTypes/GtfMacros.h>

#include "TrianglePlugin.h"
#include "GLES20TriangleRenderer.h"

#ifdef WIN32
#include "DirectX11TriangleRenderer.h"
#endif // WIN32

#include <string.h>

#include <gtf/scdr/sdk/WidgetRendererDescriptor.h>

static TriangleDesc widgetDescriptors;

// SCDR definitions of the custom properties. This should be the same properties as
// defined in TriangleDesc, TriangleProperties
static const gtf::scdr::sdk::WidgetRendererProperty WidgetProperties[] = 
{
    { 
        "color",                  // GTF/Studio name of the property
        NULL,                     // GTF/Studio name of the feature
        "Color",                  // SCDR internal name of the property.
        GtfTypeId::eTypeId_color, // GTF type of the property
        1                         // Number of elements in property
    }
};

TrianglePlugin::TrianglePlugin()
    : sdk(NULL)
    , registration(NULL)
    , initialized(false)
{
}

TrianglePlugin::~TrianglePlugin()
{
}

void TrianglePlugin::startup(GtfPluginLoader& loader_)
{
    // Initialize the GTF messenger service and set this class as listener
    GtfMsgExtReceiver::Init(loader_);

    if (GetMessenger())
    {
        GetMessenger()->SubscribeSysMsg(*this);
    }

    initialized = true;
}

void TrianglePlugin::shutdown()
{
    // Unsubscribe the messenger
    GetMessenger()->UnsubscribeSysMsg(*this);
}

uint16_t TrianglePlugin::HandleRunlevel(uint16_t const runlevel_, int8_t const direction_)
{
    uint16_t const myRunlevel = 1;
    uint16_t nextRunLevel = MAX(myRunlevel, runlevel_);

    if ((direction_ > 0)) // Startup runlevels
    {
        if (runlevel_ == myRunlevel)
        {
            // Publish the interface of the custom widget, GTF will handle this interface
            GetMessenger()->PublishInterface(
                static_cast<uint32_t>(DescriptorProvider::cm_gtfInterfaceId),
                static_cast<uint32_t>(DescriptorProvider::cm_gtfInterfaceVersion),
                0,
                static_cast<DescriptorProvider*>(&widgetDescriptors));

            // Next runlevel requested by the custom widget
            nextRunLevel = GTF_RUNLEVEL_RENDERER_PREPARE;
        }
        if (runlevel_ == GTF_RUNLEVEL_RENDERER_PREPARE)
        {
            // Register the custom renderer at the SCDR
            registerRenderer();
        }
    }
    else if ((direction_ < 0) && (true == initialized)) // Shutdown runlevels
    {
        if (runlevel_ == GTF_RUNLEVEL_RENDERER_PREPARE)
        {
            if (registration) 
            {
                // Unregister the renderer at the SCDR
                sdk->unregisterRenderer(registration);
                registration = NULL;
            }
            shutdown();
            initialized = false;
        }
    }

    return nextRunLevel;
}

void TrianglePlugin::HandleInterface(uint32_t const interfaceId_, uint32_t const version_, uint16_t const, void* const object_)
{
    if (object_)
    {
        // Handle the TypeManager interface to register our types at the GTF
        if ((interfaceId_ == static_cast<uint32_t>(GtfTypeManager::cm_gtfInterfaceId)) &&
            (version_ == static_cast<uint32_t>(GtfTypeManager::cm_gtfInterfaceVersion)))
        {
            // Resolve the types defined in the class TriangleDesc
            widgetDescriptors.ResolveTypes(static_cast<GtfTypeManager*>(object_));
        }

        // Handle the SCDR SDK interface
        if ((interfaceId_ == static_cast<uint32_t>(gtf::scdr::sdk::SDK::cm_gtfInterfaceId)) &&
            (version_ == static_cast<uint32_t>(gtf::scdr::sdk::SDK::cm_gtfInterfaceVersion)))
        {
            // Just remember the pointer. This interface is used to register the custom renderer at 
            // the SCDR
            sdk = static_cast<gtf::scdr::sdk::SDK*>(object_);
        }
    }
}

void TrianglePlugin::registerRenderer()
{
    // Register the custom renderer at the SCDR
    GTF_OSAL_ASSERT(sdk);

    if (NULL == sdk)
    {
        // There is no interface to register the custom renderer at the SCDR
        // The interface should be revived by HandleInterface
        return;
    }

    // For Properties provided by the SCDR/GTF, the SCDR defines sets of pre defined properties.
    // Defined are:
    //  * the default properties (x, y, width, height, is visible)
    //  * properties of the default widget features (Scaling, Rotation, Coloration, ....)
    //  * internal properties, which are not explicitly visible in EB GUIDE Studio like the framebuffer,
    //    and the view projection matrix.
    gtf::scdr::sdk::SystemWidgetPropertySets::Value propertySets[] = 
    {
        gtf::scdr::sdk::SystemWidgetPropertySets::Full2D
    };

    gtf::scdr::sdk::SystemWidgetProperty::Value properties[64];
    gtf::scdr::sdk::WidgetRendererProperty      renderProperties[64];

    // Expand the property sets to a list of single properties
    const uint32_t propertiesWritten = sdk->resolveSystemWidgetPropertySets(propertySets, 
        ARRAY_SIZE(propertySets), properties, ARRAY_SIZE(properties));

    // Resolve the expanded properties. 
    uint32_t renderPropertiesWritten = sdk->resolveSystemWidgetProperties(
        properties, propertiesWritten, renderProperties, 
        ARRAY_SIZE(renderProperties));

    // Add the custom properties to the list
    memcpy(renderProperties + renderPropertiesWritten, WidgetProperties, sizeof(WidgetProperties));
    renderPropertiesWritten += ARRAY_SIZE(WidgetProperties);

    // Assemble the descriptor to create the renderer
    gtf::scdr::sdk::WidgetRendererDescriptor wrdesc;

    // Set the create and destroy callbacks depending on the currently used graphics api
    if (gtf::scdr::sdk::GraphicsApi::OpenGLES20 == sdk->getGraphicsApi())
    {
        wrdesc.functions.create = GLES20TriangleRenderer::createIntance;
        wrdesc.functions.destroy = GLES20TriangleRenderer::destroyInstance;
        wrdesc.widgetRenderDataSize = GLES20TriangleRenderer::RenderDataSize; // Size of custom widget memory. 
                                                                              // A memory block of this size is provided by the SCDR
                                                                              // to store per widget temporary data. The size is defined by
                                                                              // the custom renderer.
    }

#ifdef WIN32
    if (gtf::scdr::sdk::GraphicsApi::DirectX11 == sdk->getGraphicsApi())
    {
        wrdesc.functions.create = DirectX11TriangleRenderer::createIntance;
        wrdesc.functions.destroy = DirectX11TriangleRenderer::destroyInstance;
        wrdesc.widgetRenderDataSize = DirectX11TriangleRenderer::RenderDataSize;
    }
#endif // WIN32

    wrdesc.functionUserData     = NULL;                     // An additional void pointer for custom data. 
    wrdesc.properties.count     = renderPropertiesWritten;  // Number of properties
    wrdesc.properties.values    = renderProperties;         // The property
    wrdesc.widgetName           = TriangleDesc::WidgetName; // The name of the custom widget

    // Register the new renderer. The GTF/SCDR will create an instance of the renderer by calling
    // the provided create and destroy callbacks.
    registration = sdk->registerRenderer(wrdesc);
}

// Handle a GTF message
void TrianglePlugin::HandleMessage(const GtfMsgHandle)
{
}

//GTF plugin interface
GTF_PLUGIN_SO_SYMBOL GtfPlugin * GTF_CREATE_PLUGIN_SYMBOL(uint32_t argc, char** argv)
{
    GTF_UNUSED_PARAM(argc);
    GTF_UNUSED_PARAM(argv);
    return GtfPluginFactory<TrianglePlugin>::create();
}

GTF_PLUGIN_SO_SYMBOL void GTF_DESTROY_PLUGIN_SYMBOL(GtfPlugin* const plugin)
{
    GtfPluginFactory<TrianglePlugin>::destroy(plugin);
}
